# This only tests the TokenCredentials test case, since the
# CertificateCredentials would be mocked out anyway.
# Namely:
# - timing out of the token
# - creating multiple tokens for different topics

import pytest
from freezegun import freeze_time

from apns23.credentials import TokenCredentials

TOPIC = 'com.example.first_app'


@pytest.fixture
def token_credentials():
    return TokenCredentials('test/eckey.pem', '1QBCDJ9RST', '3Z24IP123A')


def test_token_expiration(token_credentials):
    # As long as the token lifetime hasn't elapsed, this should work. To
    # be really careful, we should check how much time has elapsed to
    # know if it fail. But, either way, we'd have to come up with a good
    # lifetime for future tests...

    with freeze_time('2012-01-14'):
        expiring_header = token_credentials.get_authorization_header(TOPIC)

    new_header = token_credentials.get_authorization_header(TOPIC)
    assert expiring_header != new_header



import time
import jwt

from hyper import HTTP20Connection
from hyper.tls import init_context

DEFAULT_TOKEN_LIFETIME = 3600
DEFAULT_TOKEN_ENCRYPTION_ALGORITHM = 'ES256'


# Abstract Base class. This should not be instantiated directly.
class Credentials(object):
    def __init__(self, ssl_context=None):
        self.__ssl_context = ssl_context

    # Creates a connection with the credentials, if available or necessary.
    def create_connection(self, server, port, proto, proxy_host=None, proxy_port=None):
        # self.__ssl_context may be none, and that's fine.
        return HTTP20Connection(server, port, ssl_context=self.__ssl_context, force_proto=proto or 'h2',
                                secure=True, proxy_host=proxy_host, proxy_port=proxy_port)

    def get_authorization_header(self, topic):
        return None


# Credentials subclass for certificate authentication
class CertificateCredentials(Credentials):
    def __init__(self, cert_file=None, password=None, cert_chain=None):
        ssl_context = init_context(cert=cert_file, cert_password=password)
        if cert_chain:
            ssl_context.load_cert_chain(cert_chain)
        super(CertificateCredentials, self).__init__(ssl_context)


# Credentials subclass for JWT token based authentication
class TokenCredentials(Credentials):
    def __init__(self, auth_key_path, auth_key_id, team_id,
                 encryption_algorithm=DEFAULT_TOKEN_ENCRYPTION_ALGORITHM,
                 token_lifetime=DEFAULT_TOKEN_LIFETIME):
        self.__auth_key = self._get_signing_key(auth_key_path)
        self.__auth_key_id = auth_key_id
        self.__team_id = team_id
        self.__encryption_algorithm = encryption_algorithm
        self.__token_lifetime = token_lifetime

        # Dictionary of {topic: (issue time, ascii decoded token)}
        self.__jwt_token = None

        # Use the default constructor because we don't have an SSL context
        super(TokenCredentials, self).__init__()

    def get_authorization_header(self, topic):
        token = self._get_or_create_topic_token(topic)
        return 'bearer %s' % token

    @staticmethod
    def _is_expired_token(issue_date):
        return time.time() > issue_date + DEFAULT_TOKEN_LIFETIME

    @staticmethod
    def _get_signing_key(key_path):
        secret = ''
        if key_path:
            with open(key_path) as f:
                secret = f.read()
        return secret

    def _get_or_create_topic_token(self, topic):
        # dict of topic to issue date and JWT token
        token_pair = self.__jwt_token
        if token_pair is None or self._is_expired_token(token_pair[0]):
            # Create a new token
            issued_at = time.time()
            token_dict = {
                'iss': self.__team_id,
                'iat': issued_at,
            }
            headers = {
                'alg': self.__encryption_algorithm,
                'kid': self.__auth_key_id,
            }
            jwt_token = jwt.encode(token_dict, self.__auth_key,
                                   algorithm=self.__encryption_algorithm,
                                   headers=headers).decode('ascii')

            # Cache JWT token for later use. One JWT token per connection.
            # https://developer.apple.com/documentation/usernotifications/setting_up_a_remote_notification_server/establishing_a_token-based_connection_to_apns
            self.__jwt_token = (issued_at, jwt_token)
            return jwt_token
        else:
            return token_pair[1]



MAX_PAYLOAD_SIZE = 4096


class PayloadAlert(object):
    def __init__(self, title=None, title_localized_key=None, title_localized_args=None,
                 body=None, body_localized_key=None, body_localized_args=None,
                 action_localized_key=None, action=None,
                 launch_image=None):
        self.title = title
        self.title_localized_key = title_localized_key
        self.title_localized_args = title_localized_args
        self.body = body
        self.body_localized_key = body_localized_key
        self.body_localized_args = body_localized_args
        self.action_localized_key = action_localized_key
        self.action = action
        self.launch_image = launch_image

    def dict(self):
        result = {}

        if self.title:
            result['title'] = self.title
        if self.title_localized_key:
            result['title-loc-key'] = self.title_localized_key
        if self.title_localized_args:
            result['title-loc-args'] = self.title_localized_args

        if self.body:
            result['body'] = self.body
        if self.body_localized_key:
            result['loc-key'] = self.body_localized_key
        if self.body_localized_args:
            result['loc-args'] = self.body_localized_args

        if self.action_localized_key:
            result['action-loc-key'] = self.action_localized_key
        if self.action:
            result['action'] = self.action

        if self.launch_image:
            result['launch-image'] = self.launch_image

        return result


class Payload(object):
    def __init__(self, alert=None, badge=None, sound=None,
                 content_available=False, mutable_content=False,
                 category=None, url_args=None, custom=None, thread_id=None):
        self.alert = alert
        self.badge = badge
        self.sound = sound
        self.content_available = content_available
        self.category = category
        self.url_args = url_args
        self.custom = custom
        self.mutable_content = mutable_content
        self.thread_id = thread_id

    def dict(self):
        result = {
            'aps': {}
        }
        if self.alert is not None:
            if isinstance(self.alert, PayloadAlert):
                result['aps']['alert'] = self.alert.dict()
            else:
                result['aps']['alert'] = self.alert
        if self.badge is not None:
            result['aps']['badge'] = self.badge
        if self.sound is not None:
            result['aps']['sound'] = self.sound
        if self.content_available:
            result['aps']['content-available'] = 1
        else:
            result['aps']['content-available'] = 0
        if self.mutable_content:
            result['aps']['mutable-content'] = 1
        if self.thread_id is not None:
            result['aps']['thread-id'] = self.thread_id
        if self.category is not None:
            result['aps']['category'] = self.category
        if self.url_args is not None:
            result['aps']['url-args'] = self.url_args
        if self.custom is not None:
            result.update(self.custom)

        return result



import contextlib

try:
    from unittest.mock import MagicMock, Mock, patch
except ImportError:
    from mock import MagicMock, Mock, patch

import pytest

from apns23.client import APNsClient, CONCURRENT_STREAMS_SAFETY_MAXIMUM, Notification
from apns23.errors import ConnectionFailed
from apns23.payload import Payload

TOPIC = 'com.example.App'


@pytest.fixture(scope='session')
def tokens():
    return ['%064x' % i for i in range(1000)]


@pytest.fixture(scope='session')
def notifications(tokens):
    payload = Payload(alert='Test alert')
    return [Notification(token=token, payload=payload) for token in tokens]


@patch('apns23.credentials.init_context')
@pytest.fixture
def client(mock_connection):
    with patch('apns23.credentials.HTTP20Connection') as mock_connection_constructor:
        mock_connection_constructor.return_value = mock_connection
        return APNsClient(credentials=None)


@pytest.fixture
def mock_connection():
    mock_connection = MagicMock()
    mock_connection.__max_open_streams = 0
    mock_connection.__open_streams = 0
    mock_connection.__mock_results = None
    mock_connection.__next_stream_id = 0

    @contextlib.contextmanager
    def mock_get_response(stream_id):
        mock_connection.__open_streams -= 1
        if mock_connection.__mock_results:
            reason = mock_connection.__mock_results[stream_id]
            response = Mock(status=200 if reason == 'Success' else 400)
            response.read.return_value = ('{"reason": "%s"}' % reason).encode('utf-8')
            yield response
        else:
            yield Mock(status=200)

    def mock_request(*_args):
        mock_connection.__open_streams += 1
        mock_connection.__max_open_streams = max(mock_connection.__open_streams, mock_connection.__max_open_streams)

        stream_id = mock_connection.__next_stream_id
        mock_connection.__next_stream_id += 1
        return stream_id

    mock_connection.get_response.side_effect = mock_get_response
    mock_connection.request.side_effect = mock_request
    mock_connection._conn.__enter__.return_value = mock_connection._conn
    mock_connection._conn.remote_settings.max_concurrent_streams = 500

    return mock_connection


def test_connect_establishes_connection(client, mock_connection):
    client.connect()
    mock_connection.connect.assert_called_once_with()


def test_connect_retries_failed_connection(client, mock_connection):
    mock_connection.connect.side_effect = [RuntimeError, RuntimeError, None]
    client.connect()
    assert mock_connection.connect.call_count == 3


def test_connect_stops_on_reaching_max_retries(client, mock_connection):
    mock_connection.connect.side_effect = [RuntimeError] * 4
    with pytest.raises(ConnectionFailed):
        client.connect()

    assert mock_connection.connect.call_count == 3


def test_send_empty_batch_does_nothing(client, mock_connection):
    client.send_notification_batch([], TOPIC)
    assert mock_connection.request.call_count == 0


def test_send_notification_batch_returns_results_in_order(client, mock_connection, tokens, notifications):
    results = client.send_notification_batch(notifications, TOPIC)
    expected_results = {token: 'Success' for token in tokens}
    assert results == expected_results


def test_send_notification_batch_respects_max_concurrent_streams_from_server(client, mock_connection, tokens,
                                                                             notifications):
    client.send_notification_batch(notifications, TOPIC)
    assert mock_connection.__max_open_streams == 500


def test_send_notification_batch_overrides_server_max_concurrent_streams_if_too_large(client, mock_connection, tokens,
                                                                                      notifications):
    mock_connection._conn.remote_settings.max_concurrent_streams = 5000
    client.send_notification_batch(notifications, TOPIC)
    assert mock_connection.__max_open_streams == CONCURRENT_STREAMS_SAFETY_MAXIMUM


def test_send_notification_batch_overrides_server_max_concurrent_streams_if_too_small(client, mock_connection, tokens,
                                                                                      notifications):
    mock_connection._conn.remote_settings.max_concurrent_streams = 0
    client.send_notification_batch(notifications, TOPIC)
    assert mock_connection.__max_open_streams == 1


def test_send_notification_batch_reports_different_results(client, mock_connection, tokens,
                                                           notifications):
    mock_connection.__mock_results = (
            ['BadDeviceToken'] * 1000 + ['Success'] * 1000 + ['DeviceTokenNotForTopic'] * 2000 +
            ['Success'] * 1000 + ['BadDeviceToken'] * 500 + ['PayloadTooLarge'] * 4500
    )
    results = client.send_notification_batch(notifications, TOPIC)
    expected_results = dict(zip(tokens, mock_connection.__mock_results))
    assert results == expected_results



#!/usr/bin/env python

from setuptools import setup

dependencies = [
    'hyper>=0.7',
    'PyJWT>=1.4.0',
    'cryptography>=1.7.2',
    'enum-compat==0.0.3',
]


setup(
    name='apns23',
    version='1.0.1',
    packages=['apns23'],
    install_requires=dependencies,
    extras_require={
        "tests": [
            'freezegun',
            'pytest',
            'mock',
        ],
    },
    url='https://github.com/VinayGValsaraj/PyAPNs23',
    license='MIT',
    author='Vinay Valsaraj',
    author_email='vinaygvalsaraj@gmail.com',
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.6',
        'Topic :: Software Development :: Libraries',
    ],
    description='A python library for interacting with the Apple Push Notification Service via HTTP/2 protocol'
)



class APNsException(Exception):
    pass


class ConnectionFailed(APNsException):
    """There was an error connecting to APNs."""


class InternalException(APNsException):
    """This exception should not be raised. If it is, please report this as a bug."""
    pass


class BadPayloadException(APNsException):
    """Something bad with the payload."""
    pass


class BadCollapseId(BadPayloadException):
    """The collapse identifier exceeds the maximum allowed size"""
    pass


class BadDeviceToken(APNsException):
    """The specified device token was bad.
    Verify that the request contains a valid token and that the token matches the environment."""
    pass


class BadExpirationDate(BadPayloadException):
    """The apns-expiration value is bad."""
    pass


class BadMessageId(InternalException):
    """The apns-id value is bad."""
    pass


class BadPriority(InternalException):
    """The apns-priority value is bad."""
    pass


class BadTopic(BadPayloadException):
    """The apns-topic was invalid."""
    pass


class DeviceTokenNotForTopic(APNsException):
    """The device token does not match the specified topic."""


class DuplicateHeaders(InternalException):
    """One or more headers were repeated."""
    pass


class IdleTimeout(APNsException):
    """Idle time out."""
    pass


class MissingDeviceToken(APNsException):
    """The device token is not specified in the request :path.
    Verify that the :path header contains the device token."""
    pass


class MissingTopic(BadPayloadException):
    """The apns-topic header of the request was not specified and was required.
    The apns-topic header is mandatory when the client is connected using a certificate
    that supports multiple topics."""
    pass


class PayloadEmpty(BadPayloadException):
    """The message payload was empty."""
    pass


class TopicDisallowed(BadPayloadException):
    """Pushing to this topic is not allowed."""
    pass


class BadCertificate(APNsException):
    """The certificate was bad."""
    pass


class BadCertificateEnvironment(APNsException):
    """The client certificate was for the wrong environment."""
    pass


class ExpiredProviderToken(APNsException):
    """The provider token is stale and a new token should be generated."""
    pass


class Forbidden(APNsException):
    """The specified action is not allowed."""
    pass


class InvalidProviderToken(APNsException):
    """The provider token is not valid or the token signature could not be verified."""
    pass


class MissingProviderToken(APNsException):
    """No provider certificate was used to connect to APNs and Authorization header was missing or no provider token
    was specified. """
    pass


class BadPath(APNsException):
    """The request contained a bad :path value."""
    pass


class MethodNotAllowed(InternalException):
    """The specified :method was not POST."""
    pass


class Unregistered(APNsException):
    """The device token is inactive for the specified topic."""

    def __init__(self, timestamp=None):
        super(Unregistered, self).__init__()

        self.timestamp = timestamp


class PayloadTooLarge(BadPayloadException):
    """The message payload was too large. The maximum payload size is 4096 bytes."""
    pass


class TooManyProviderTokenUpdates(APNsException):
    """The provider token is being updated too often."""
    pass


class TooManyRequests(APNsException):
    """Too many requests were made consecutively to the same device token."""
    pass


class InternalServerError(APNsException):
    """An internal server error occurred."""
    pass


class ServiceUnavailable(APNsException):
    """The service is unavailable."""
    pass


class Shutdown(APNsException):
    """The server is shutting down."""
    pass


def exception_class_for_reason(reason):
    return {
        'BadCollapseId': BadCollapseId,
        'BadDeviceToken': BadDeviceToken,
        'BadExpirationDate': BadExpirationDate,
        'BadMessageId': BadMessageId,
        'BadPriority': BadPriority,
        'BadTopic': BadTopic,
        'DeviceTokenNotForTopic': DeviceTokenNotForTopic,
        'DuplicateHeaders': DuplicateHeaders,
        'IdleTimeout': IdleTimeout,
        'MissingDeviceToken': MissingDeviceToken,
        'MissingTopic': MissingTopic,
        'PayloadEmpty': PayloadEmpty,
        'TopicDisallowed': TopicDisallowed,
        'BadCertificate': BadCertificate,
        'BadCertificateEnvironment': BadCertificateEnvironment,
        'ExpiredProviderToken': ExpiredProviderToken,
        'Forbidden': Forbidden,
        'InvalidProviderToken': InvalidProviderToken,
        'MissingProviderToken': MissingProviderToken,
        'BadPath': BadPath,
        'MethodNotAllowed': MethodNotAllowed,
        'Unregistered': Unregistered,
        'PayloadTooLarge': PayloadTooLarge,
        'TooManyProviderTokenUpdates': TooManyProviderTokenUpdates,
        'TooManyRequests': TooManyRequests,
        'InternalServerError': InternalServerError,
        'ServiceUnavailable': ServiceUnavailable,
        'Shutdown': Shutdown,
    }[reason]



import collections
import json
import logging
import time
import weakref
from enum import Enum
from threading import Thread

from .errors import ConnectionFailed, exception_class_for_reason

# We don't generally need to know about the Credentials subclasses except to
# keep the old API, where APNsClient took a cert_file
from .credentials import CertificateCredentials


class NotificationPriority(Enum):
    Immediate = '10'
    Delayed = '5'


RequestStream = collections.namedtuple('RequestStream', ['stream_id', 'token'])
Notification = collections.namedtuple('Notification', ['token', 'payload'])

DEFAULT_APNS_PRIORITY = NotificationPriority.Immediate
CONCURRENT_STREAMS_SAFETY_MAXIMUM = 1000
MAX_CONNECTION_RETRIES = 3

logger = logging.getLogger(__name__)


class APNsClient(object):
    SANDBOX_SERVER = 'api.sandbox.push.apple.com'
    LIVE_SERVER = 'api.push.apple.com'

    DEFAULT_PORT = 443
    ALTERNATIVE_PORT = 2197

    def __init__(self, credentials, use_sandbox=False, use_alternative_port=False, proto=None, json_encoder=None,
                 password=None, proxy_host=None, proxy_port=None, heartbeat_period=None):
        if credentials is None or isinstance(credentials, str):
            self.__credentials = CertificateCredentials(credentials, password)
        else:
            self.__credentials = credentials
        self._init_connection(use_sandbox, use_alternative_port, proto, proxy_host, proxy_port)

        if heartbeat_period:
            self._start_heartbeat(heartbeat_period)

        self.__json_encoder = json_encoder
        self.__max_concurrent_streams = None
        self.__previous_server_max_concurrent_streams = None

    def _init_connection(self, use_sandbox, use_alternative_port, proto, proxy_host, proxy_port):
        server = self.SANDBOX_SERVER if use_sandbox else self.LIVE_SERVER
        port = self.ALTERNATIVE_PORT if use_alternative_port else self.DEFAULT_PORT
        self._connection = self.__credentials.create_connection(server, port, proto, proxy_host, proxy_port)

    def _start_heartbeat(self, heartbeat_period):
        conn_ref = weakref.ref(self._connection)

        def watchdog():
            while True:
                conn = conn_ref()
                if conn is None:
                    break

                conn.ping('-' * 8)
                time.sleep(heartbeat_period)

        thread = Thread(target=watchdog)
        thread.setDaemon(True)
        thread.start()

    def send_notification(self, token_hex, notification, topic=None, priority=NotificationPriority.Immediate,
                          expiration=None, collapse_id=None):
        stream_id = self.send_notification_async(token_hex, notification, topic, priority, expiration, collapse_id)
        result = self.get_notification_result(stream_id)
        if result != 'Success':
            if isinstance(result, tuple):
                reason, info = result
                raise exception_class_for_reason(reason)(info)
            else:
                raise exception_class_for_reason(result)

    def send_notification_async(self, token_hex, notification, topic=None, priority=NotificationPriority.Immediate,
                                expiration=None, collapse_id=None):
        json_str = json.dumps(notification.dict(), cls=self.__json_encoder, ensure_ascii=False, separators=(',', ':'))
        json_payload = json_str.encode('utf-8')

        headers = {
            'apns-push-type': 'alert'
        }

        if topic is not None:
            headers['apns-topic'] = topic

        if priority != DEFAULT_APNS_PRIORITY:
            headers['apns-priority'] = priority.value

        if expiration is not None:
            headers['apns-expiration'] = '%d' % expiration

        auth_header = self.__credentials.get_authorization_header(topic)
        if auth_header is not None:
            headers['authorization'] = auth_header

        if collapse_id is not None:
            headers['apns-collapse-id'] = collapse_id

        url = '/3/device/{}'.format(token_hex)
        stream_id = self._connection.request('POST', url, json_payload, headers)
        return stream_id

    def get_notification_result(self, stream_id):
        """
        Get result for specified stream
        The function returns: 'Success' or 'failure reason' or ('Unregistered', timestamp)
        """
        with self._connection.get_response(stream_id) as response:
            if response.status == 200:
                return 'Success'
            else:
                raw_data = response.read().decode('utf-8')
                data = json.loads(raw_data)
                if response.status == 410:
                    return data['reason'], data['timestamp']
                else:
                    return data['reason']

    def send_notification_batch(self, notifications, topic=None, priority=NotificationPriority.Immediate,
                                expiration=None, collapse_id=None):
        """
        Send a notification to a list of tokens in batch. Instead of sending a synchronous request
        for each token, send multiple requests concurrently. This is done on the same connection,
        using HTTP/2 streams (one request per stream).

        APNs allows many streams simultaneously, but the number of streams can vary depending on
        server load. This method reads the SETTINGS frame sent by the server to figure out the
        maximum number of concurrent streams. Typically, APNs reports a maximum of 500.

        The function returns a dictionary mapping each token to its result. The result is "Success"
        if the token was sent successfully, or the string returned by APNs in the 'reason' field of
        the response, if the token generated an error.
        """
        notification_iterator = iter(notifications)
        next_notification = next(notification_iterator, None)
        # Make sure we're connected to APNs, so that we receive and process the server's SETTINGS
        # frame before starting to send notifications.
        self.connect()

        results = {}
        open_streams = collections.deque()
        # Loop on the tokens, sending as many requests as possible concurrently to APNs.
        # When reaching the maximum concurrent streams limit, wait for a response before sending
        # another request.
        while len(open_streams) > 0 or next_notification is not None:
            # Update the max_concurrent_streams on every iteration since a SETTINGS frame can be
            # sent by the server at any time.
            self.update_max_concurrent_streams()
            if self.should_send_notification(next_notification, open_streams):
                logger.info('Sending to token %s', next_notification.token)
                stream_id = self.send_notification_async(next_notification.token, next_notification.payload, topic,
                                                         priority, expiration, collapse_id)
                open_streams.append(RequestStream(stream_id, next_notification.token))

                next_notification = next(notification_iterator, None)
                if next_notification is None:
                    # No tokens remaining. Proceed to get results for pending requests.
                    logger.info('Finished sending all tokens, waiting for pending requests.')
            else:
                # We have at least one request waiting for response (otherwise we would have either
                # sent new requests or exited the while loop.) Wait for the first outstanding stream
                # to return a response.
                pending_stream = open_streams.popleft()
                result = self.get_notification_result(pending_stream.stream_id)
                logger.info('Got response for %s: %s', pending_stream.token, result)
                results[pending_stream.token] = result

        return results

    def should_send_notification(self, notification, open_streams):
        return notification is not None and len(open_streams) < self.__max_concurrent_streams

    def update_max_concurrent_streams(self):
        # Get the max_concurrent_streams setting returned by the server.
        # The max_concurrent_streams value is saved in the H2Connection instance that must be
        # accessed using a with statement in order to acquire a lock.
        # pylint: disable=protected-access
        with self._connection._conn as connection:
            max_concurrent_streams = connection.remote_settings.max_concurrent_streams

        if max_concurrent_streams == self.__previous_server_max_concurrent_streams:
            # The server hasn't issued an updated SETTINGS frame.
            return

        self.__previous_server_max_concurrent_streams = max_concurrent_streams
        # Handle and log unexpected values sent by APNs, just in case.
        if max_concurrent_streams > CONCURRENT_STREAMS_SAFETY_MAXIMUM:
            logger.warning('APNs max_concurrent_streams too high (%s), resorting to default maximum (%s)',
                           max_concurrent_streams, CONCURRENT_STREAMS_SAFETY_MAXIMUM)
            self.__max_concurrent_streams = CONCURRENT_STREAMS_SAFETY_MAXIMUM
        elif max_concurrent_streams < 1:
            logger.warning('APNs reported max_concurrent_streams less than 1 (%s), using value of 1',
                           max_concurrent_streams)
            self.__max_concurrent_streams = 1
        else:
            logger.info('APNs set max_concurrent_streams to %s', max_concurrent_streams)
            self.__max_concurrent_streams = max_concurrent_streams

    def connect(self):
        """
        Establish a connection to APNs. If already connected, the function does nothing. If the
        connection fails, the function retries up to MAX_CONNECTION_RETRIES times.
        """
        retries = 0
        while retries < MAX_CONNECTION_RETRIES:
            try:
                self._connection.connect()
                logger.info('Connected to APNs')
                return
            except Exception:  # pylint: disable=broad-except
                # close the connnection, otherwise next connect() call would do nothing
                self._connection.close()
                retries += 1
                logger.exception('Failed connecting to APNs (attempt %s of %s)', retries, MAX_CONNECTION_RETRIES)

        raise ConnectionFailed()
